# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from copy import copy
import kollacli.i18n as u

from kollacli.common.inventory import Inventory
from kollacli.common.utils import check_arg
from kollacli.common.utils import safe_decode


class HostApi(object):

    def host_add(self, hostnames):
        # type: (List[str]) -> None
        """Add hosts to the inventory

        :param hostnames: list of strings
        """
        check_arg(hostnames, u._('Host names'), list)
        hostnames = safe_decode(hostnames)

        inventory = Inventory.load()
        any_changed = False
        for hostname in hostnames:
            changed = inventory.add_host(hostname)
            if changed:
                any_changed = True
        if any_changed:
            Inventory.save(inventory)

    def host_remove(self, hostnames):
        # type: (List[str]) -> None
        """Remove hosts from the inventory

        :param hostnames: list of strings
        """
        check_arg(hostnames, u._('Host names'), list)
        hostnames = safe_decode(hostnames)

        inventory = Inventory.load()
        any_changed = False
        for hostname in hostnames:
            changed = inventory.remove_host(hostname)
            if changed:
                any_changed = True
        if any_changed:
            Inventory.save(inventory)

    def host_get_all(self):
        # type: () -> List[Host]
        """Get all hosts in the inventory

        :return: Hosts
        :rtype: list of Host objects
        """
        inventory = Inventory.load()
        hosts = []
        host_groups = inventory.get_host_groups()
        for hostname, groupnames in host_groups.items():
            hosts.append(Host(hostname, groupnames))
        return hosts

    def host_get(self, hostnames):
        # type: (List[str]) -> List[Host]
        """Get selected hosts in the inventory

        :param hostnames: list of strings
        :return: hosts
        :rtype: list of Host objects
        """
        check_arg(hostnames, u._('Host names'), list)
        hostnames = safe_decode(hostnames)
        inventory = Inventory.load()
        inventory.validate_hostnames(hostnames)

        hosts = []
        host_groups = inventory.get_host_groups()
        for hostname in hostnames:
            hosts.append(Host(hostname, host_groups[hostname]))
        return hosts

    def host_ssh_check(self, hostnames):
        # type: (List[str]) -> Dict[str,Dict[str,object]]
        """Check hosts for ssh connectivity

        Check status is a dictionary of form:
            - {hostname: {
              'success':<True|False>,
              'msg':message_string},
              ...
              }

        :param hostnames: list of strings
        :return: check status
        :rtype: dictionary
        """
        check_arg(hostnames, u._('Host names'), list)
        inventory = Inventory.load()
        hostnames = safe_decode(hostnames)
        inventory.validate_hostnames(hostnames)
        summary = inventory.ssh_check_hosts(hostnames)
        return summary

    def host_setup(self, hosts_info):
        # type: (Dict[str,Dict[str,object]]) -> None
        """Setup multiple hosts for ssh access

        hosts_info is a dictionary of form:
            - {hostname': {
              'password': password
              'uname': user_name},
              ...
              }

        The uname entry is optional.

        :param hosts_info: dictionary
        """
        check_arg(hosts_info, u._('Hosts info'), dict)
        inventory = Inventory.load()
        inventory.validate_hostnames(hosts_info.keys())
        inventory.setup_hosts(hosts_info)


class Host(object):
    """Host"""

    def __init__(self, hostname, groupnames):
        # type: (str, List[str]) -> None
        self.name = hostname
        self._groupnames = groupnames

    def get_name(self):
        # type: () -> str
        """Get name

        :return: host name
        :rtype: string
        """
        return self.name

    def get_groups(self):
        # type: () -> List[str]
        """Get names of the groups associated with this host

        :return: group names
        :rtype: list of strings

        Note: If the groups associated with this host change after this
        host is fetched, the host must be re-fetched to reflect those
        changes.
        """
        return copy(self._groupnames)
